/*
Copyright (C) 1996-2001 Id Software, Inc.
Copyright (C) 2013-2014 Baker

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// r_common.c

#include "quakedef.h"

	
//
// view origin
//
vec3_t vup /* qasm */, vpn /* qasm */, vright /* qasm */; // ASM uses for sound only I think
vec3_t r_origin /* qasm */;
mplane_t frustum[4];

float r_fovx, r_fovy;

entity_t *currententity;
vec3_t modelorg  /* qasm */;

refdef_t r_refdef /* qasm */;

//
// screen size info
//


//
// refresh flags
//

// Put that into r_light.c



frame_render_t frame; // Baker: Doesn't really belong here.  Need to create an renderer common file?
level_info_t level; // Baker: Likewise, doesn't really belong here.




void R_Init (void)
{
	Cmd_AddCommands (R_Init);

		
	Sky_Init (); //johnfitz

	R_InitParticles (); // After R_InitTextures ?

	R_Init_Local (); // Runs either the rest of software or GL initialization
}


void R_NewMap (void)
{
	int		i;

	for (i=0 ; i < 256 ; i++)
		d_lightstylevalue[i] = 264;		// normal light value

// clear out efrags in case the level hasn't been reloaded
// FIXME: is this one short?

	for (i = 0 ; i < cl.worldmodel->numleafs ; i++)
		cl.worldmodel->leafs[i].efrags = NULL;

	R_ClearParticles ();

	View_NewMap ();			// Baker: Clear viewshifts

#ifdef SUPPORTS_CUTSCENE_PROTECTION
	Cvar_Clear_Untrusted ();	// Baker: Ensure that cut-scenes and demos restore to correct values by firewalling cvar sets.
#endif // SUPPORTS_CUTSCENE_PROTECTION

	Sky_NewMap (); //johnfitz -- skybox in worldspawn


	R_NewMap_Local ();
}


/*
====================
R_TimeRefresh_f

For program optimization
====================
*/
void R_TimeRefresh_f (void)
{
	int			i;
	float		start, stop, elapsedtime;

	if (cls.state != ca_connected)
	{
		Con_Printf("Not connected to a server\n");
		return;
	}

	start = System_DoubleTime ();
	for (i=0 ; i < 128 ; i++)
	{
		VID_BeginRendering (&clx, &cly, &clwidth, &clheight);
		r_refdef.viewangles[1] = i/128.0*360.0;

		R_RenderView ();
		VID_EndRendering ();
	}

#ifdef GLQUAKE_RENDERER_SUPPORT
	eglFinish ();
#endif // GLQUAKE_RENDERER_SUPPPORT

	stop = System_DoubleTime ();
	elapsedtime = stop-start;
	Con_Printf ("%f seconds (%f fps)\n", time, 128/elapsedtime);
}


void R_SetLiquidAlpha (void)
{
	if (!level.water_vis_known && frame.has_abovewater && frame.has_underwater)
	{
		// Baker: Don't let this scenario lock us into a false reading.
		// Although this would be an extremely hard scenario to generate, would take an incredibly well placed saved game
		// and I tried hard to stand somewhere a save game cause this problem and couldn't despite my best efforts.

		if (!frame.nearwaterportal && !r_novis.value)
		{
			Con_DPrintf ("AUTO WATER VIS:  Level is vised!\n");
			level.water_vis_known = true;
			level.water_vis = true;
		}
	}

	if (r_novis.value)											frame.liquid_alpha = true;  // r_novis 1
	else if (level.water_vis_known && level.water_vis)			frame.liquid_alpha = true;  // Known to be watervised
	else if (level.water_vis_known && !level.water_vis)			frame.liquid_alpha = false; // Known to be not watervised
	else if (frame.has_abovewater && frame.has_underwater)		frame.liquid_alpha = true;	// Weird situation almost impossible
	else														frame.liquid_alpha = false;	 // Vis not known yet, but no water brushes in scene

	if (frame.liquid_alpha)
	{
		frame.wateralpha	= CLAMP(0, r_wateralpha.value, 1.0);
		frame.slimealpha	= CLAMP(0, r_slimealpha.value, 1.0);
		frame.lavaalpha		= CLAMP(0, r_lavaalpha.value, 1.0);
	} else frame.wateralpha = frame.slimealpha = frame.lavaalpha = 1;
}



// Baker: Look for cvars and commands in common

/*
===============
R_TextureAnimation -- johnfitz -- added "frame" param to eliminate use of "currententity" global

Returns the proper texture for a given time and base texture
===============
*/
texture_t *R_TextureAnimation (texture_t *base, int frame)
{
	int		relative;
	int		count;

	if (frame)
	{
		if (base->alternate_anims)
			base = base->alternate_anims;
	}

	if (!base->anim_total)
		return base;

	relative = (int)(cl.ctime*10) % base->anim_total;

	count = 0;
	while (base->anim_min > relative || base->anim_max <= relative)
	{
		base = base->anim_next;
		if (!base)
			Host_Error ("R_TextureAnimation: broken cycle");
		if (++count > 100)
			Host_Error ("R_TextureAnimation: infinite cycle");
	}

	return base;
}

